import math

from .rangierlib.ablauf import rangier_vorwaerts
from .rangierlib.ausparken import ausparkvorgang

def angle_to_distance(action, car_rad):
    if action[1] in ["rechts", "links"]:
        # Wenn 2 pi = pi * 2 * d sind, dann
        distance = action[2] * car_rad
    else:
        distance = action[2]
        
    return (action[0], action[1], distance)

def translate_action(action):
    front_back = "vorwärts" if action[0] == "rückwärts" else "rückwärts"
    return (front_back, action[1], action[2])

class Rangierer():
    def __init__(self, w=.19, f=.33, b=.05, r=.658, p=.05, l=.45, verbose=False):
        self.verbose = verbose
        self.w = w
        self.f = f
        self.b = b
        self.r = r
        self.p = p
        self.l = l
        # Init:
        self.x_auto = -(l-b)
        self.y_auto = -(p+w)
        self.x_VA = 0
        self.y_VA = -(w/2+p)
        self.x_HA = -l
        self.y_HA = self.y_VA
        
        self.car_pos = (self.x_auto, self.y_auto)
        self.p_VA = (self.x_VA, self.y_VA)
        self.p_HA = (self.x_HA, self.y_HA)
        
        self.actions=[]
        final_x = self.calculate_actions()
        
        self.einpark_actions = self.actions[::-1]
        self.einpark_actions = [angle_to_distance(translate_action(action),self.r) for action in self.einpark_actions]
        self.einpark_actions.insert(0, ("vorwärts", "geradeaus", final_x))
        self.auspark_actions = [angle_to_distance(action, self.r) for action in self.actions]
        
    def calculate_actions(self):
        car_pos_nach_rangieren, alpha_nach_rangieren = rangier_vorwaerts(
            self.car_pos, 0, self.r, self.p_VA, self.p_HA, self.f, self.w, self.b, self.actions, self.verbose)

        # print(
        #     f"car_pos_nach_rangieren: {car_pos_nach_rangieren}, alpha_nach_rangieren: {alpha_nach_rangieren}")
        x_final, y_final = ausparkvorgang(
            car_pos_nach_rangieren, alpha_nach_rangieren, self.r, self.p_VA, self.f, self.w, self.b, 0, self.actions)

        # print(f"x_final: {x_final}")
        return x_final
        
    



if __name__ == "__main__":
    """
    Dieser Test geht einmal den kompletten Algorithmus durch, wie er dann später auch in Wirklichkeit durchgeführt wird. 
    
    Das Auto hat einen Parkplatz erkannt und steht nun mit seiner Hinterachse genau an der Kante, wo das Vorderauto aufhört.
    """
    # hinterauto_pos = (7.844, 32.447)

    car_rad = 65.88
    f = 33
    w = 19
    b = 5
    
    p = 5.306
    l = 49.0
    
    rangierer = Rangierer(w, f, b, car_rad, p, l)
    print("Einparkactions:")
    print(rangierer.einpark_actions)
    print("Ausparkactions:")
    print(rangierer.auspark_actions)
    # actions = []
    
    # # Werte sollten alle mit dem Modell übereinstimmen. 
    # print(f"car_pos: ({rangierer.x_auto}, {rangierer.y_auto})")
    # print(f"p_VA: ({rangierer.x_VA}, {rangierer.y_VA})")
    # print(f"p_HA: ({rangierer.x_HA}, {rangierer.y_HA})")
    
    # car_pos = (rangierer.x_auto, rangierer.y_auto)
    # p_VA = (rangierer.x_VA, rangierer.y_VA)
    # p_HA = (rangierer.x_HA, rangierer.y_HA)
    # car_pos_nach_rangieren, alpha_nach_rangieren = rangier_vorwaerts(car_pos, 0, car_rad, p_VA, p_HA, f, w, b, actions)
    
    # print(f"car_pos_nach_rangieren: {car_pos_nach_rangieren}, alpha_nach_rangieren: {alpha_nach_rangieren}")
    # x_final, y_final = ausparkvorgang(car_pos_nach_rangieren, alpha_nach_rangieren, car_rad, p_VA, f, w, b, 0, actions)
    
    # print(f"x_final: {x_final}")
    # print(actions)
    



# from fsolvetest import *
from scipy.optimize.minpack import fsolve
from .fsolver import alpha_verschiebung_hinten_links, dist_func_to_minimize, ecke_vorne_rechts, ecke_hinten_links, center_of_circle, alpha_verschiebung_vorne_rechts, radius_ecke_hinten_links, radius_ecke_vorne_rechts
import math
"""
Bei dem Algorithmus fangen wir wieder mal beim ausparken an.
Das Ziel ist es, mit der Mitte der Hinterachse den Ursprung des Koordinatensystems zu erreichen, welcher genau dort liegt, wo man aus der Parklücke perfekt waagerecht draußen ist. 

Es wird zunächst ein Ausparkvorgang beschrieben. Um statt auszuparken einzuparken muss einfach bei jeder Parkoperation das Vorzeichen vom Winkel und die Fahrtrichtung umgekehrt werden. Außerdem natürlich die gesamte Liste der Operationen.

Wir fangen an, zugemauert in einer winzigen Parklücke, nach hinten hin sehr dicht an dem Hinterauto. Es gibt zunächst zwei Operationen, die sich rekursiv aufrufen und in eine Liste von Anweisungen die später ausgeführt werden sollen ihren Fahrtwinkel + Lenkwinkel + Fahrrichtung schreiben.

Die beiden Operationen heißen: RangierVorwärts und RangierRückwärts. 

RangierVorwärts: 
Schlage voll nach links ein und prüfe, ob wir aus der Lücke rauskämen:
- Falls ja: BREAK: Leite den Ausparkvorgang ein
- Falls nein: Fahr soweit wie möglich nach vorne, RangierRückwärts

RangierRückwärts:
Schlage voll nach rechts ein und berechne, wie weit (alpha) man fahren müsste, um beim nach vorne fahren aus der Lücke kommen zu können.
Berechne, ob es möglich ist, so weit zu fahren, ohne mit der hinteren linken Ecke ans hinterauto anzustoßen. 
- Falls ja: Fahre genau so weit und rangiere vorwärts (sollte dann sofort durchpassen zum Ausparkvorgang)
- Falls nein: Fahre so weit wie es möglich ist, und rangiere vorwärts. 


"""


def rangier_vorwaerts(car_pos, alpha, car_rad, p_VA, p_HA, f, w, b, actions, verbose = False):
    if verbose:
        print(
            f"\nRangiere Vorwärts mit car_pos: {car_pos}, alpha: {alpha} ({math.degrees(alpha)}°)")

    outer_front_car_rad = radius_ecke_vorne_rechts(car_rad, f, w)
    x_vorne_rechts_new, y_vorne_rechts_new, alpha_vorne_rechts_new = pos_und_alpha_vordere_rechte_ecke_so_weit_wie_möglich(
        car_rad, outer_front_car_rad, car_pos, alpha, p_VA[0])

    upper_circle_center = center_of_circle(
        car_rad, alpha, car_pos, "forward")
    # TODO: Das ganze hier doch als Klasse schreiben, die Parameter nerven doch hart...
    # print(y_vorne_rechts_new, p_VA[1])
    if y_vorne_rechts_new >= p_VA[1]:  # Hier muss evtl statt y_vr y_auto hin!
        if verbose:
            print("Yey, man kann ausparken, mache ich jetzt auch.")
        return car_pos, alpha
        ...  # Fahr raus, break!
    else:
        alpha_new = alpha_vorne_rechts_new - \
            alpha_verschiebung_vorne_rechts(car_rad, f, w)
        x_new = upper_circle_center[0] + car_rad * math.cos(alpha_new)
        y_new = upper_circle_center[1] + car_rad * math.sin(alpha_new)
        alpha_new_my_system = alpha_new - 1.5*math.pi
        delta_alpha = alpha_new - 1.5 * math.pi - alpha
        if verbose:
            print(
                f"Nach dem vorwärts rangieren wird das Auto den Winkel: {alpha_new_my_system} ({math.degrees(alpha_new_my_system)}°) haben. Es hat sich demnach um {delta_alpha} ({math.degrees(delta_alpha)}°) bewegt. Es steht nun an der Position ({x_new}, {y_new}).")

        actions.append(("vorwärts", "links", delta_alpha))
        return rangier_rueckwaerts((x_new, y_new), alpha_new_my_system,
                                   car_rad, p_VA, p_HA, f, w, b, actions)


def rangier_rueckwaerts(car_pos, alpha, car_rad, p_VA, p_HA, f, w, b, actions, verbose = False):
    """
    1. Berechne, wie weit man rückwärts fahren müsste, um vorwärts rauskommen zu können.
    """
    if verbose:
        print(
            f"\nRangiere rückwärts mit car_pos: {car_pos}, alpha: {alpha}({math.degrees(alpha)}°)")

    lower_circle_center = center_of_circle(car_rad, alpha, car_pos, "backward")
    # print(f"Der lower circle center liegt bei: {lower_circle_center}")
    outer_front_car_rad = radius_ecke_vorne_rechts(car_rad, f, w)

    function_to_minimize = dist_func_to_minimize(
        car_rad, outer_front_car_rad, p_VA, alpha, car_pos)
    solutions = fsolve(function_to_minimize, 0)
    # print(f"solutions: {solutions}")
    # print(f"Das Auto müsste {math.degrees(solutions[0])}° nach hinten fahren, damit es vorne gut rauskommt.")

    # todo: Ohne den Wert fährt er gerade so nicht weit genug zurück...
    delta_alpha = solutions[0] + 0.005
    alpha_new = alpha + delta_alpha
    absolute_alpha_new = alpha_new + .5*math.pi

    # print(math.degrees(absolute_alpha_new))
    x_new = lower_circle_center[0] + car_rad * math.cos(absolute_alpha_new)
    y_new = lower_circle_center[1] + car_rad * math.sin(absolute_alpha_new)

    potentielle_hintere_linke_ecke = ecke_hinten_links(
        x_new, y_new, alpha_new, b, w)

    # Man kann so weit zurücksetzen wie gewünscht. Beende die Rekursion
    if potentielle_hintere_linke_ecke[0] > p_HA[0]:
        if verbose:
            print(
                f"Nach dem Rückwärts rangieren wird das Auto den Winkel: {alpha_new} ({math.degrees(alpha_new)}°) haben. Es hat sich demnach um {delta_alpha} ({math.degrees(delta_alpha)}°) bewegt. Es steht nun an der Position: ({x_new}, {y_new})")
        actions.append(("rückwärts", "rechts", delta_alpha))
        return rangier_vorwaerts((x_new, y_new), alpha_new, car_rad, p_VA, p_HA, f, w, b, actions)
    if verbose:
        print(
            f"Leider kann ich nicht so weit zurückfahren, wie ich gerne würde {potentielle_hintere_linke_ecke[1]} < {p_HA[0]}...")
    # Fahre nur soweit zurück, bis das p_HA[0] erreicht wird.
    # TODO: Die Position der potentiellen hinteren linken Ecke validieren!
    outer_back_car_rad = radius_ecke_hinten_links(car_rad, b, w)
    x_hinten_links_new, y_hinten_links_new, alpha_hinten_links_new = pos_und_alpha_hintere_linke_ecke_so_weit_wie_möglich(
        car_rad, outer_back_car_rad, car_pos, alpha, p_HA[0])

    alpha_new = alpha_hinten_links_new + \
        alpha_verschiebung_hinten_links(car_rad, b, w)
    x_new = lower_circle_center[0] + car_rad * math.cos(alpha_new)
    y_new = lower_circle_center[1] + car_rad * math.sin(alpha_new)

    alpha_new_my_system = alpha_new - 0.5 * math.pi
    delta_alpha = alpha_new_my_system - alpha

    if verbose:
        print(
            f"Nach dem Rückwärts SOWEIT WIE MÖGLICH rangieren wird das Auto den Winkel: {alpha_new_my_system} ({math.degrees(alpha_new_my_system)}°) haben. Es hat sich demnach um {delta_alpha} ({math.degrees(delta_alpha)}°) bewegt. Es steht nun an der Position: ({x_new}, {y_new})")
    actions.append(("rückwärts", "rechts", delta_alpha))
    return rangier_vorwaerts((x_new, y_new), alpha_new_my_system,
                             car_rad, p_VA, p_HA, f, w, b, actions)

def pos_und_alpha_hintere_linke_ecke_so_weit_wie_möglich(car_rad, outer_back_car_rad, car_pos, initial_alpha, x_HA):
    """
    Berechnet den x und y Wert und den absoluten Alpha Winkel der hinteren linken Ecke des Autos, wenn es so weit nach hinten fährt, wie es kann (bis es beim x-Wert des hinteren Autos ist.)
    """
    lower_circle_center = center_of_circle(car_rad, initial_alpha, car_pos, "backward")
    alpha = math.acos((x_HA - lower_circle_center[0]) / outer_back_car_rad)
    
    # alpha = 2*math.pi - alpha
    y = lower_circle_center[1] + outer_back_car_rad * math.sin(alpha)
    x = lower_circle_center[0] + outer_back_car_rad * math.cos(alpha)
    # y ist der Ort, an dem die vordere rechte Ecke des Autos landet. Das zugehörige x ist

    return x, y, alpha
    
    
    

def pos_und_alpha_vordere_rechte_ecke_so_weit_wie_möglich(car_rad, outer_front_car_rad, car_pos, initial_alpha, x_VA):
    """ 
    Berechnet, den x und y Wert von der vorderen rechten Ecke des Autos, wenn es so weit nach vorne fährt, wie es kann (bis es beim x-Wert des vorderen Autos ist.)
    Wenn dann der zugehörige y-Wert höher ist, als die obere Beschränkung des Autos, kann zum Ausparken geleitet werden.
    """

    upper_circle_center = center_of_circle(
        car_rad, initial_alpha, car_pos, "forward")

    alpha = math.acos(
        (x_VA - upper_circle_center[0]) / outer_front_car_rad)
    # Das Alpha ist für den oberen Halbkreis, wir brauchen aber das für den unteren:
    alpha = 2*math.pi - alpha
    y = upper_circle_center[1] + outer_front_car_rad * math.sin(alpha)
    x = upper_circle_center[0] + outer_front_car_rad * math.cos(alpha)
    # y ist der Ort, an dem die vordere rechte Ecke des Autos landet. Das zugehörige x ist
    
    return x, y, alpha
    

if __name__ == "__main__":
    """
    Realistisches Beispiel zum Test von delta_y_und_alpha_je_nach_x_nach_vorne
    Auto steht anfangs bei (5.238,10.732)
    Radius des Autos liegt bei: 65.88
    VA_Ecke liegt bei: (50.176, 32.447)
    Das Auto ist um 20° eingelenkt (0.1111111111111111 * pi).
    Das Auto ist von der Hinterachse aus noch 33cm nach vorne und 19 cm breit.

    Erst müsste die Ecke vorne Rechts berechnet werden. In Fusion liegt diese bei: (39.497, 13.092)
    """
    car_pos = (5.238, 10.732)
    target_pos = (50.176, 32.447)
    hinterauto_pos = (5.37, 32.447)
    # hinterauto_pos = (7.844, 32.447)

    initial_alpha = 0.1111111111*math.pi
    car_rad = 65.88
    outer_front_car_rad = 82.2287
    f = 33
    w = 19
    b = 5
    
    x, y, alpha = pos_und_alpha_vordere_rechte_ecke_so_weit_wie_möglich(
        car_rad, outer_front_car_rad, car_pos, initial_alpha, target_pos[0])
    
    """
    x = xm + r * cos a
    a = acos((xm - x)/r)
    y = ym + r * sin a
    Aus der Berechnung von x besorgt man sich das alpha
    und dann kann man das y aus der anderen holen.
    
    PROBLEM: Ich kriege nur die Obere Ecke dessen zu fassen was ich will.
    Aber der gesuchte x-Wert wird ja zwei mal getroffen. Wie sorge ich dafür, dass ich den richtigen finde??
    """
    
    # print(f"Das Alpha nach neuer Rechnung: {alpha} / {math.degrees(alpha)}° resultiert in x:{x}, y: {y}")
    actions = []
    rangier_vorwaerts(car_pos, initial_alpha, car_rad, target_pos, hinterauto_pos, f, w, b, actions)
    """
    Nach dem ersten Rangieren sollte in unserem Beispiel das Auto an der Position 
    (17.121, 16.463) stehen. Das ist korrekt. Außerdem sollte es den Winkel 31.43° haben (also um 11° nach vorne links gefahren sein).
    
    Das Rückwärts rangieren würde berechnen, dass das Auto nun 3.3° rückwärts fahren sollte, um aus der Lücke rauszukommen.
    """
    
    """
    Um das Modell beim Rückwärts fahren richtig zu nerven sollte hinten ein Auto auf Höhe x = 4.5 stehen.
    """
    
    
    """
    Dann wird berechnet, wie weit es noch nach vorne fahren sollte (auf der x-Achse.)
    """
    # x_distanz_nach_vorne = target_pos[0] - x_vr
    
    # dy, dalpha = delta_y_und_alpha_je_nach_x_nach_vorne(x_distanz_nach_vorne, car_rad)
    
    # print(f"Das auto müsste noch {x_distanz_nach_vorne}cm nach vorne (x) fahren. Dazu würde es den Winkel {dalpha} abfahren und dabei {dy} an Höhe gewinnen. Die Endposition wäre also bei: {(car_pos[0]+x_distanz_nach_vorne, car_pos[1]+dy)}")

    
    
    



import math


def circle_intersection(x1, y1, r1, x2, y2, r2):
    '''
    @summary: calculates intersection points of two circles
    @param circle1: tuple(x,y,radius)
    @param circle2: tuple(x,y,radius)
    @result: tuple of intersection points (which are (x,y) tuple)
    '''
    # http://stackoverflow.com/a/3349134/798588
    dx, dy = x2-x1, y2-y1
    d = math.sqrt(dx*dx+dy*dy)
    if d > r1+r2:
        # print("#1")
        return [1] # no solutions, the circles are separate
    if d < abs(r1-r2):
        # print("#2")
        return [2]  # no solutions because one circle is contained within the other
    if d == 0 and r1 == r2:
        # print("#3")
        return [3]  # circles are coincident and there are an infinite number of solutions

    a = (r1*r1-r2*r2+d*d)/(2*d)
    h = math.sqrt(r1*r1-a*a)
    xm = x1 + a*dx/d
    ym = y1 + a*dy/d
    xs1 = xm + h*dy/d
    xs2 = xm - h*dy/d
    ys1 = ym - h*dx/d
    ys2 = ym + h*dx/d

    return (xs1, ys1), (xs2, ys2)




if __name__ == "__main__":
    
    """
    Auto ist in der Position, wo es nach links rausfahren könnte, die Frage ist, wie weit es nach links rausfährt. 
    Es sollte so lange rausfahren, bis der rechte Wendekreis des Autos genau an einer Stelle sich mit dem dem Kreis mit Radius w/2 mit Mittelpunkt p_VA befindet.
    p_VA und w/2 sind fix. Genauso r vom Wendekreis     
    """
    
    p_VA = (50.176, 32.447)
    w = 19
    car_rad = 65.88
    
    """
    p1: Das Auto sollte noch zwei Schnittpunkte mit dem Kreis haben
    """
    p1 = (49.684, -32.757)
    
    circle_intersection(p_VA[0], p_VA[1], w/2, p1[0], p1[1], car_rad)
    # Das ist korrekt
    
    """
    p2: Das Auto sollte gerade noch so zwei schnittpunkte haben:
    """
    p2 = (64.549, -22.157)
    circle_intersection(p_VA[0], p_VA[1], w/2, p2[0], p2[1], car_rad)




import numpy as np
from .circle_intersections import circle_intersection
import math

from .fsolver import center_of_circle


def delta_x_und_y_um_rechts_gerade_zu_lenken(alpha, car_rad):
    """
    Bekommt ein Alpha, wie das Auto momentan steht und berechnet, wie weit es noch in der y-Achse nach oben geht, wenn es gerade gelenkt wird. 
    """
    # es geht von alpha + 90° bis 90°
    # y = sin a * r
    y1 = math.sin(alpha+0.5*math.pi) * car_rad
    y2 = math.sin(0.5*math.pi) * car_rad
    yd = y2-y1
    
    x1 = math.cos(alpha+0.5*math.pi) * car_rad
    x2 = math.cos(0.5*math.pi) * car_rad
    xd = x2-x1

    return xd,yd

def ausparkvorgang(car_pos, alpha, car_rad, p_VA, f, w, b, y_final, actions):
    """
    Das Auto steht nun an einer Position, bei der es voll nach links eingeschlagen an dem vorderen Auto vorbeikommen würde. 
    Der Ausparkvorgang sieht wie folgt aus:
    
    1. Fahre so lange voll links eingeschlagen nach vorne, bis der rechte Wendekreis zu p_VA mindestens/genau einen Abstand von w/2 hat.
    1.5 Berechne, wie viel y Wir durch das geradelenken bei gegebenem Winkel gemacht werden
    2. Fahre dann so lange gerade, bis man dann wenn man nach rechts einschlägt auf der richtigen Höhe landen wird.
    3. Lenke gerade. 
    Jetzt kann es sein, dass der Winkel des Autos zu steil ist. 
    - in dem Fall: Verkackt, kann jemand anders machen. 
    
    
    Wie kriegen wir es nun hin, dass dieser Abstand jetzt richtig ist?
    - Ich brauche ja diesmal keinen Abstand zu einem Punkt, sondern einem Kreis. 
    - Es geht ja schon auch darum, solange zu drehen, bis man nur noch einen Schnittpunkt mit dem 
    
    zu 1. 
    Zunächst brauchen wir die den Mittelpunkt vom rechten Wendekreis in Abhängigkeit vom Drehwinkel -> gibt es ja schon. center_of_circle("backward")
    
    Erhöhe dann solange den den Drehwinkel (maximal bis 90° so), bis der return von circle_intersection == 2 ist. 
    
    
    """
    
    # 1. Finde den Winkel, der nach links abgefahren werden muss (append to actions)
    new_car_pos, new_alpha = finde_links_fahr_winkel_beim_ausparken(car_pos, alpha, car_rad, p_VA, w)
    
    # TODO: Handle None (wird momentan nen error schmeißen, wenn nix rauskommt)
    
    
    # Berechne die Position die das Auto haben wird nachdem der Winkel gefahren ist
    actions.append(("vorwärts", "links", new_alpha-alpha))
    
    # 2. Berechne, wie weit nach x und y gefahren würde, beim gerade lenken in der Rechtskurve.
    dx, dy = delta_x_und_y_um_rechts_gerade_zu_lenken(new_alpha, car_rad)
    
    # 3. Berechne aus der letzten Position und dem geradelenken, wie weit noch geradeaus gefahren werden muss (welches +x resultiert)
    y_missing = y_final - (new_car_pos[1] + dy)
    
    if y_missing < 0:
        print("Leider kann in diese Parklücke nicht gefahren werden.")
        return
        
    hypothenuse_geradeaus_zu_fahren = y_missing/math.sin(new_alpha)
    x_missing = y_missing/math.tan(new_alpha)
    
    # 4. Füge die letzten Anweisungen hinzu und fertig ist der Ausparkprozess
    actions.append(("vorwärts", "geradeaus", hypothenuse_geradeaus_zu_fahren))
    actions.append(("vorwärts", "rechts", new_alpha))
    x_final = new_car_pos[0] + dx + x_missing
    
    return (x_final, y_final)
    
    
    ...


def finde_links_fahr_winkel_beim_ausparken(car_pos, alpha, car_rad, p_VA, w):

    lower_circle_mid = center_of_circle(car_rad, alpha, car_pos, "backward")

    first_intersection = circle_intersection(
        p_VA[0], p_VA[1], w/2, lower_circle_mid[0], lower_circle_mid[1], car_rad)
    
    upper_circle_center = center_of_circle(
        car_rad, alpha, car_pos, "forward")

    # print(f"first intersection {first_intersection}")
    if len(first_intersection) != 2:
        print("Dangerzone! Es sollte am Anfang schon zwei Überschneidungen geben")
    for possible_alpha in np.linspace(alpha, 0.5*math.pi, 100):

        """
        1. calculate new car pos with new alpha
        2. calculate new lower_circle_mid with new alpha and car pos
        """
        x_new = upper_circle_center[0] + car_rad * \
            math.cos(1.5*math.pi + possible_alpha)
        y_new = upper_circle_center[1] + car_rad * \
            math.sin(1.5*math.pi + possible_alpha)
        car_pos = (x_new, y_new)

        lower_circle_mid = center_of_circle(
            car_rad, possible_alpha, car_pos, "backward")
        # print(f"upper circle mid: {upper_circle_center} car pos: {car_pos} lower circle mid: {lower_circle_mid}")

        intersection = circle_intersection(
            p_VA[0], p_VA[1], w/2, lower_circle_mid[0], lower_circle_mid[1], car_rad)
        if len(intersection) == 1:
            # print(f"FOUND IT!{possible_alpha}")
            return car_pos, possible_alpha
        
    print("Konnte nicht den richtigen Winkel zum Ausparken finde. Ab jetzt wirds holprig (ich garantiere für nichts mehr).")
    ...


if __name__ == "__main__":
    p_VA = (50.176, 32.447)
    w = 19
    car_rad = 65.88
    initial_car_pos = (17.437, 24.8)
    
    initial_alpha = math.radians(33.6)
    print(f"initial alpha: {initial_alpha}")


    finde_links_fahr_winkel_beim_ausparken(initial_car_pos, initial_alpha, car_rad, p_VA, w)



from scipy.optimize import fsolve
import math
 

def ecke_vorne_rechts(x,y,a, f = .33, w = .19):
    """ Ergibt mit Modellbeispiel (39.497, 13.091) ist korrekt."""
    xn = x + f * math.cos(a) + (w/2) * math.sin(a)
    yn = y - (w/2) * math.cos(a) + f * math.sin(a)
    return (xn, yn)


def ecke_hinten_links(x, y, a, b=.05, w=.19):
    """ Ergibt mit Modellbeispiel (1.942, 19.641) ist korrekt."""
    xn = x - b * math.cos(a) - (w/2) * math.sin(a)
    yn = y + (w/2) * math.cos(a) - b * math.sin(a) 
    return (xn, yn)


def radius_ecke_vorne_rechts(r = .6588, f=.33, w=.19):
    """ Ergibt mit Modellbeispiel 82.2869 ist korrekt."""
    pos_ecke_vorne_rechts = ecke_vorne_rechts(0,0,0, f, w)
    pos_auto_drehkreis_mittelpunkt = (0, r)
    return math.dist(pos_ecke_vorne_rechts, pos_auto_drehkreis_mittelpunkt)


def radius_ecke_hinten_links(r=.6588, b=.05, w=.19):
    """ Ergibt mit Modellbeispiel 75.546 ist korrekt."""
    pos_ecke_hinten_links = ecke_hinten_links(0, 0, 0, b, w)
    pos_auto_drehkreis_mittelpunkt = (0, -r)
    return math.dist(pos_ecke_hinten_links, pos_auto_drehkreis_mittelpunkt)


def alpha_verschiebung_vorne_rechts(r, f, w):
    """ Ergibt mit Modellbeispiel 23.643 ist korrekt."""
    pos_ecke_vorne_rechts = ecke_vorne_rechts(0, 0, 0, f, w)
    pos_auto_drehkreis_mittelpunkt = (0, r)
    radius_vorne_rechts = radius_ecke_vorne_rechts(r,f,w)
    alpha_vorne_rechts = math.acos(
        (pos_ecke_vorne_rechts[0] - pos_auto_drehkreis_mittelpunkt[0]) / radius_vorne_rechts)
    alpha_vorne_rechts = 2*math.pi - alpha_vorne_rechts
    return alpha_vorne_rechts - 1.5*math.pi



def alpha_verschiebung_hinten_links(r, b, w):
    """ Ergibt mit Modellbeispiel -5.1 ist korrekt."""
    pos_ecke_hinten_links = ecke_hinten_links(0, 0, 0, b, w)
    pos_auto_drehkreis_mittelpunkt = (0, r)
    radius_hinten_links = radius_ecke_hinten_links(r, b, w)
    alpha_hinten_links = math.acos(
        (pos_ecke_hinten_links[0] - pos_auto_drehkreis_mittelpunkt[0]) / radius_hinten_links)
    alpha_hinten_links = 2*math.pi - alpha_hinten_links
    return alpha_hinten_links - 1.5*math.pi


def center_of_circle(r, initial_a, p_car, direction):
    if direction == "forward":
        initial_a = initial_a + 1.5 * math.pi
    else:
        initial_a = initial_a + 0.5 * math.pi

    x_mid = p_car[0] - r * math.cos(initial_a)
    y_mid = p_car[1] - r * math.sin(initial_a)
    return x_mid, y_mid


def position_after_rotation(a, r, initial_a = 0, p_car = (0,0), direction="forward"):
    """
    Berechnet die Position auf einer Kreisbahn nach einer Rotation.
    Die Rotationsrichtung ist nach vorne links.
    """
    
    # Erstmal den Mittelpunkt von dem Kreis ausrechnen, aus dem initialen Winkel und p_car
    x_mid, y_mid = center_of_circle(r, initial_a, p_car, direction)
    # print(f"center: {(x_mid, y_mid)}")
    
    if direction == "forward":
        initial_a = initial_a + 1.5* math.pi
    else:
        initial_a = initial_a + 0.5*math.pi

    x = r * math.cos(a + initial_a) + x_mid
    y = r * math.sin(a + initial_a) + y_mid
    
    return x,y


def dist_func_to_minimize(r_car, r_outer_edge_car, p_target, initial_a = 0, p_car = (0,0)):
    """
    Bekommt die Werte der drei Parameter und gibt die verkettete
    Funktion von pos und dist zurück, die nur noch a als Parameter braucht. 
    """
    def return_func(a):
        p_car_after_rotation = position_after_rotation(a, r_car, initial_a, p_car, "backward")
        center_of_upper_circle = center_of_circle(
            r_car, initial_a + a, p_car_after_rotation, "forward")
        d = math.dist(center_of_upper_circle, p_target) - r_outer_edge_car
        return d
    return return_func


"""
Realistisches Beispiel:
Auto steht anfangs bei (5.238,10.732)
Radius des Autos liegt bei: 65.88
Radius der vorderen rechten Ecke liegt bei 82.2287
Das auto ist bereits um 20° (0.1111111111111111 * pi) gedreht.

Der Punkt den wir erreichen wollen liegt bei: (50.176, 32.447)

Frage: Wie weit müssen wir noch nach hinten drehen, bis wir nach vorne mit der Schnauze vorbeikommen?
"""

if __name__ == "__main__":
    car_pos = (5.238, 10.732)
    target_pos = (50.176, 32.447)
    initial_alpha = 0.1111111111*math.pi
    car_rad = 65.88
    outer_front_car_rad = 82.2287

    rotated_alpha = initial_alpha
    """
    Erstes Händisches Beispiel: 
    Siehe in der Fusion Datei Test_nach_hinten_fahren
    Dort wird der Sachverhalt modelliert wo wir landen, wenn das Auto bereits 20° eingedreht ist und weitere 20° nach hinten fährt. 
    Zunächst ist dann der Mittelpunkt vom Auto Bei (-14.576,-0.708) und der Mittelpunkt vom oberen Kreis bei (-56.28, 48.99). 
    """
    pos_after_back_rotation = position_after_rotation(
        rotated_alpha, car_rad, initial_alpha, car_pos, "backward")
    print(f"Position after back rotation: {pos_after_back_rotation}")
    initial_uper_circle_middle = center_of_circle(
        car_rad, initial_alpha + rotated_alpha, pos_after_back_rotation, "forward")
    print(f"initial mid of upper circle should be: {initial_uper_circle_middle}")

    """
    Erstes Optimierungsbeispiel:
    Wenn wir nun unsere Funktion optimieren möchten, dann sollte das Auto noch ca. 2.5° nach hinten gedreht werden, bevor man mit der Schnauze vorbei fahren könnte.
    Dies soll der minimizer nun selbst rausfinden:
    """

    example_function = dist_func_to_minimize(
        car_rad, outer_front_car_rad, target_pos, initial_alpha, car_pos)
    solutions = fsolve(example_function, 0)
    print(f"Das Auto muss noch {math.degrees(solutions[0])}° nach hinten fahren.")
